(* Copyright (C) 1992, Digital Equipment Corporation                         *)
(* All rights reserved.                                                      *)
(* See the file COPYRIGHT for a full description.                            *)
(*                                                                           *)
(* Last modified on Tue Nov 23 16:46:29 PST 1993 by kalsow                   *)
(*      modified on Thu May  6 16:25:50 PDT 1993 by muller                   *)

MODULE DragonInt;

IMPORT Word;

PROCEDURE FixSize (VAR a: T) =
  BEGIN
    a.s := NUMBER (a.w^);
    WHILE a.s > 0 AND a.w [a.s - 1] = 0 DO
      DEC (a.s); END;
  END FixSize;
    
PROCEDURE copy (READONLY a: T): T =
  VAR res: T;
  BEGIN
    res.s := a.s;
    res.w := NEW (WA, a.s);
    res.w^ := SUBARRAY (a.w^, 0, a.s);
    RETURN res;
  END copy;
 
PROCEDURE add (READONLY a,b: T): T =
  VAR res: T; carry: INTEGER; x: Word.T;
  BEGIN
    IF a.s < b.s THEN
      RETURN add(b,a);  END;

    carry := 0;
    res.w := NEW (WA, a.s + 1);

    FOR i := 0 TO b.s - 1 DO
      x := Word.Plus (Word.Plus (a.w[i], b.w[i]), carry);
      res.w[i] := Word.And (x, ButTop);
      carry := Word.RightShift (x, Used); END;

    FOR i := b.s TO a.s - 1 DO
      x := Word.Plus (a.w[i], carry);
      res.w[i] := Word.And (x, ButTop);
      carry := Word.RightShift (x, Used); END;

    res.w [a.s] := carry;
    FixSize (res);
    RETURN res;
  END add;

PROCEDURE diff (READONLY a,b: T): T =
  VAR res: T; borrow: INTEGER; x: Word.T;
  BEGIN
    <* ASSERT a.s >= b.s *>

    borrow := 0;
    res.w := NEW (WA, a.s);

    FOR i := 0 TO b.s - 1 DO
      x := Word.Minus (Word.Minus (a.w[i], b.w[i]), borrow);
      res.w[i] := Word.And (x, ButTop);
      borrow := Word.And (Word.RightShift (x, Used), 1); END;

    FOR i := b.s TO a.s - 1 DO
      x := Word.Minus (a.w[i], borrow);
      res.w[i] := Word.And (x, ButTop);
      borrow := Word.And (Word.RightShift (x, Used), 1); END;

    <*ASSERT borrow = 0*>

    FixSize (res);
    RETURN res;
  END diff;

PROCEDURE compare (READONLY a, b: T): [-1..1] =
  BEGIN
    IF a.s < b.s THEN
      RETURN -1;
    ELSIF a.s > b.s THEN 
      RETURN +1; 
    ELSE
      FOR i := a.s - 1 TO 0 BY -1 DO
        IF Word.LT (a.w[i], b.w[i]) THEN
          RETURN -1;
        ELSIF Word.GT (a.w[i], b.w[i]) THEN
          RETURN +1; END; END;
      RETURN 0; END;
  END compare;

PROCEDURE max (READONLY a, b: T): T =
  BEGIN
    IF compare (a, b) < 0 THEN
      RETURN copy (b);
    ELSE
      RETURN copy (a); END;
  END max;

PROCEDURE shiftb (READONLY a: T; n: INTEGER): T =
  (* to the left for positive n, to the right for negative n *)
  VAR res: T; k: INTEGER; carry: Word.T;
  BEGIN
    IF n = 0 OR a.s = 0 THEN
      RETURN copy (a); END;

    IF 0 < n THEN
      carry := 0;
      k := n DIV (Used);
      n := n MOD (Used);
      res.w := NEW (WA, a.s + k + 1);
      FOR i := 0 TO k - 1 DO
        res.w [i] := 0; END;
      FOR i := 0 TO a.s - 1 DO
        res.w [i + k] := Word.Or (Word.And (Word.LeftShift (a.w[i], n), 
                                            ButTop),
                                  carry);
        carry := Word.RightShift (a.w[i], Used - n); END;
      res.w [a.s + k] := carry;

    ELSE
      carry := 0;
      k := (-n) DIV Word.Size;
      n := (-n) MOD Word.Size;
      res.w := NEW (WA, a.s - k);
      FOR i := a.s - k - 1 TO 0 BY -1 DO
        res.w [i] := Word.Or (carry, Word.RightShift (a.w[i+k], n));
        carry := Word.And (Word.LeftShift (a.w[i+k], Used - n),
                           ButTop); END; END;

    FixSize (res);
    RETURN res;
  END shiftb;

PROCEDURE times2 (READONLY a: T): T =
  BEGIN
    RETURN shiftb (a, 1);
  END times2;

PROCEDURE timesBinPlace (VAR a: T): T =
  VAR carry, x: Word.T;
  BEGIN
    carry := 0;

    FOR i := 0 TO a.s - 1 DO
      x := Word.Plus (Word.Times (a.w [i], 10), carry);
      a.w [i] := Word.And (x, ButTop);
      carry := Word.RightShift (x, Used); END;

    IF carry # 0 THEN
      IF NUMBER (a.w^) = a.s THEN
        VAR wa := NEW (WA, a.s + 1); BEGIN
          SUBARRAY (wa^, 0, a.s) := a.w^;
          a.w := wa; END; END;
      a.w [a.s] := carry;
      INC (a.s); END;

    RETURN a;
  END timesBinPlace;

PROCEDURE divideB (READONLY a: T): T =
  (* upper *)
  VAR res: T; carry: INTEGER; x: INTEGER;
  BEGIN
    res.w := NEW (WA, a.s);
    carry := 0;

    FOR i := a.s - 1 TO 0 BY -1 DO
      x := Word.Or (Word.LeftShift (carry, Used), a.w[i]);
      res.w [i] := Word.Divide (x, 10);
      carry := Word.Mod (x, 10); END;
    FixSize (res);
    IF carry # 0 THEN
      res := add (res, One); END;
    FixSize (res);
    RETURN  res;
  END divideB;

PROCEDURE divmod (READONLY a, b : T; VAR d: INTEGER): T =
  (* The div is known to be a base B digit *)
  VAR n: INTEGER; nb, n1b: T;
  BEGIN
    n := 1;
    nb := b;
    n1b := Zero;
    LOOP
      IF compare (a, nb) < 0 THEN
        d := n - 1;
        RETURN diff (a, n1b);
      ELSE
        n1b := nb;
        INC (n);
        nb := add (nb, b); END; END;
  END divmod;


BEGIN
  Zero.w := NEW (WA, 1);
  Zero.s := 0;

  One.w := NEW (WA, 1);
  One.s := 1;
  One.w [0] := 1;
END DragonInt.
