(* Copyright (C) 1992, Digital Equipment Corporation                         *)
(* All rights reserved.                                                      *)
(* See the file COPYRIGHT for a full description.                            *)
(*                                                                           *)
(* Last modified on Tue Nov 23 16:40:33 PST 1993 by kalsow                   *)
(*      modified on Fri May  7 16:13:29 PDT 1993 by muller                   *)

MODULE DragonT;

IMPORT DragonInt;
FROM DragonInt IMPORT timesBinPlace, divideB, max, shiftb, compare, add;
FROM DragonInt IMPORT times2, copy, divmod, diff;

PROCEDURE F (e: INTEGER; READONLY f: DragonInt.T;  p: INTEGER;
             cutoffMode: CutoffMode;
             cutoffPlace: INTEGER;
             VAR digits: ARRAY OF Digit;
             VAR count: CARDINAL;
             VAR exp: INTEGER) =
  VAR
    roundUpFlag := FALSE;
    U, k: INTEGER;
    R, S, Mm, Mp: DragonInt.T;
    low, high: BOOLEAN;

  PROCEDURE cutoffAdjust () =
    VAR a: INTEGER; y: DragonInt.T;
    BEGIN
      a := cutoffPlace - k;
      y := S;
      IF a >= 0 THEN 
        FOR j := 1 TO  a DO 
          y := timesBinPlace (y); END; END;
      IF a <= 0 THEN 
        FOR j := 1 TO -a DO 
          y := divideB (y); END; END;
      Mm := max (y, Mm);
      Mp := max (y, Mp);
      IF compare (Mp, y) = 0 THEN
        roundUpFlag := TRUE; END;
    END cutoffAdjust;
    
  PROCEDURE fixup () = 
    BEGIN
      IF compare (f, shiftb (DragonInt.One, p-1)) = 0 THEN
        (* Account for unequal gaps *)
        Mp := shiftb (Mp, 1);
        R := shiftb (R, 1);
        S := shiftb (S, 1); END;
      
      k := 0;
      WHILE compare (R, divideB (S)) < 0 DO
        DEC (k);
        R := timesBinPlace (R);
        Mm := timesBinPlace (Mm);
        Mp := timesBinPlace (Mp);  END;
      
      REPEAT
        WHILE compare (add (times2 (R), Mp), times2 (S)) >= 0 DO
          S := timesBinPlace (S);
          INC (k); END;
        
        (* Perform any necessary adjustment of Mm and Mp to take into account
           the formatting requirements *)
        CASE cutoffMode OF 
        | CutoffMode.normal =>      cutoffPlace := k;
        | CutoffMode.absolute =>    cutoffAdjust ();
        | CutoffMode.relative =>    INC (cutoffPlace, k); 
                                    cutoffAdjust (); END;
      UNTIL compare (add (times2 (R), Mp), times2 (S)) < 0;
    END fixup;

  BEGIN
    IF compare (f, DragonInt.Zero) = 0 THEN
      count := 1;
      exp := 0;
      digits [0] := 0;
      RETURN; END;

    count := 0;
    R := shiftb (f, MAX (e - p, 0));
    S := shiftb (DragonInt.One, MAX (0, - (e - p)));
    Mm := shiftb (DragonInt.One, MAX (e - p, 0));
    Mp := copy (Mm);
    fixup ();

    LOOP
      DEC (k);
      R := divmod (timesBinPlace (R), S, U);
      Mm := timesBinPlace (Mm);
      Mp := timesBinPlace (Mp);
      low := compare (shiftb (R, 1), Mm) < 0;
      
      WITH twoS = shiftb (S, 1) DO
        IF compare (twoS, Mp) < 0 THEN
          high := TRUE;
        ELSIF roundUpFlag THEN
          high := compare (times2 (R), diff (twoS, Mp)) >= 0;
        ELSE
          high := compare (times2 (R), diff (twoS, Mp)) > 0; END; END;

      IF count = 0 THEN
        exp := k; END;

      IF NOT low AND NOT high AND k # cutoffPlace THEN
        digits [count] := U;
        INC (count);
      ELSE
        IF low AND NOT high THEN
          digits [count] := U;
        ELSIF high AND NOT low THEN
          digits [count] := U + 1; 
        ELSE
          CASE compare (shiftb (R, 1), S) OF
          | -1 => digits [count] := U;
          | 0  => digits [count] := U  (* could be U + 1 *);
          | 1  => digits [count] := U + 1; END; END;
        INC (count);
        RETURN; END; END;
  END F;

BEGIN
END DragonT.
