(*| Copyright (C) 1994, Digital Equipment Corporation           *)
(*| All rights reserved.                                        *)
(*| See the file COPYRIGHT for a full description.              *)
(*|                                                             *)
(*| Last modified on Wed May 25 15:22:36 PDT 1994 by kalsow     *)
(*|      modified on Wed May 25 14:41:19 PDT 1994 by detlefs    *)

UNSAFE MODULE RTHeapDebug;

IMPORT RTCollector, RTHeapMap, RTutils,
       WeakRef, Wr, Fmt, Thread, Env, Scan, FloatMode, Lex;
IMPORT RefSeq, IntSeq, IntRefTbl;

FROM Stdio IMPORT stderr;

<*FATAL Wr.Failure, Thread.Alerted*>

VAR freeRefs := NEW(RefSeq.T).init(maxFree);
    n := 0;

PROCEDURE Free(r: REFANY) =
  VAR wra := NEW(REF WeakRef.T); BEGIN
    wra^ := WeakRef.FromRef(r);
    FOR i := 0 TO freeRefs.size()-1 DO
      IF freeRefs.get(i) = NIL THEN
        freeRefs.put(i, wra); RETURN
      END (* IF *)
    END (* FOR *);
    IF freeRefs.size() < maxFree THEN
      freeRefs.addhi(wra)
    ELSE
      freeRefs.put(n MOD maxFree, wra); INC(n)
    END (* IF *)
  END Free;

PROCEDURE CheckHeap() = 
  PROCEDURE WalkRefAtAddress(a: ADDRESS) =
    VAR i := LOOPHOLE(a, UNTRACED REF INTEGER)^; ra: REFANY; BEGIN
      IF done OR visited.get(i, ra) THEN RETURN END (* IF *);
      IF freeAddrs.get(i, ra) THEN
        Wr.PutText(stderr, "Path to 'free' object:\n");
        Wr.PutText(stderr,
                   "   Ref in root at address 0x" &
                   Fmt.Int(path.get(0), 16) & "...\n");
        FOR j := 1 TO path.size()-1 DO
          VAR r := LOOPHOLE(path.get(j), REFANY); BEGIN
            Wr.PutText(stderr,
                       "   Object of type " &
                       RTutils.TypecodeName(TYPECODE(r)) &
                       " at address " & Fmt.Int(path.get(j), 16) & "...\n")
          END (* BEGIN *)
        END (* FOR *);
        VAR r := LOOPHOLE(a, UNTRACED REF REFANY)^; BEGIN
          Wr.PutText(stderr,
                     "   Free object of type " &
                     RTutils.TypecodeName(TYPECODE(r)) &
                     " at address " &
                     Fmt.Int(LOOPHOLE(a, INTEGER), 16) & "...\n")
        END (* BEGIN *);
        EVAL freeAddrs.delete(i, ra);
        IF freeAddrs.size() = 0 THEN done := TRUE END (* IF *)
      ELSE
        IF path.size() = 0 THEN path.addhi(LOOPHOLE(a, INTEGER)) END (* IF *);
        VAR addr := LOOPHOLE(a, UNTRACED REF ADDRESS)^; BEGIN
          IF addr # NIL THEN
            path.addhi(i); EVAL visited.put(i, NIL);
            VAR optr := LOOPHOLE(addr - BYTESIZE(RTHeapMap.ObjectPtr),
                                 RTHeapMap.ObjectPtr);
            BEGIN
              RTHeapMap.WalkRef(optr, WalkRefAtAddress)
            END (* BEGIN *);
            EVAL path.remhi(); EVAL visited.delete(i, ra);
          END (* IF *)
        END (* BEGIN *);
        IF path.size() = 1 THEN EVAL path.remhi() END (* IF *)
      END (* IF *)
    END WalkRefAtAddress;
  VAR
    freeAddrs := NEW(IntRefTbl.Default).init();
    visited := NEW(IntRefTbl.Default).init();
    path := NEW(IntSeq.T).init();
    done: BOOLEAN := FALSE;
  BEGIN
    RTCollector.Disable();
    FOR i := 0 TO freeRefs.size()-1 DO
      VAR wr: REF WeakRef.T := freeRefs.get(i); BEGIN
        IF wr # NIL THEN
          VAR r := WeakRef.ToRef(wr^); BEGIN
            IF r # NIL THEN
              EVAL freeAddrs.put(LOOPHOLE(r, INTEGER), NIL)
            ELSE
              freeRefs.put(i, NIL)
            END (* IF *)
          END (* BEGIN *)
        END (* IF *)
      END (* BEGIN *)
    END (* FOR *);
    IF freeAddrs.size() # 0 THEN
      RTHeapMap.WalkGlobals(WalkRefAtAddress)
    END (* IF *);
    RTCollector.Enable();
  END CheckHeap;

BEGIN
  VAR mfStr := Env.Get("RTHEAPDEBUG_MAXFREE"); BEGIN
    IF mfStr # NIL THEN
      TRY 
        VAR mf := Scan.Int(mfStr); BEGIN
          IF mf > 0 THEN maxFree := mf END (* IF *)
        END (* BEGIN *)
      EXCEPT Lex.Error, FloatMode.Trap =>
      END (* TRY *)
    END (* IF *)
  END (* BEGIN *)
END RTHeapDebug.
