(*| Copyright (C) 1990, Digital Equipment Corporation       *)
(*| All rights reserved.                                    *)
(*| See the file COPYRIGHT for a full description.          *)
(*|                                                         *)
(*| Last modified on Mon May 23 09:07:04 PDT 1994 by kalsow *)

UNSAFE MODULE RTHeapInfo;

IMPORT RT0, RT0u, RTHeapRep, RTParams, RTPerfTool, RTType;
IMPORT Text, Thread, Cstring;

TYPE Closure = Thread.Closure OBJECT OVERRIDES apply := Producer END;

VAR
  drain  : RTPerfTool.Handle;
  self   : Thread.T := NIL;
  cnt    : INTEGER := 0;
  buf    : ARRAY [0..255] OF INTEGER;
  update : LONGREAL := 5.0d0;

PROCEDURE Producer (<*UNUSED*> self: Thread.Closure): REFANY =
  VAR n: INTEGER;  nTypes: RT0.Typecode := RT0u.nTypes;
  BEGIN
    LOOP
      RTHeapRep.Lock ();
        (* count the non-zero stats *)
        n := 0;
        FOR i := 0 TO nTypes-1 DO
          IF RT0u.alloc_cnts[i] # 0 THEN INC (n) END;
        END;

        (* and send'm *)
        Send (n);
        FOR i := 0 TO nTypes-1 DO
          IF RT0u.alloc_cnts[i] # 0 THEN
            Send (i);
            Send (RT0u.alloc_cnts[i]);
            RT0u.alloc_cnts[i] := 0;
            IF (RT0u.alloc_bytes[i] # 0) THEN
              Send (RT0u.alloc_bytes[i]);
              RT0u.alloc_bytes[i] := 0;
            END;
          END;
        END;
        Flush ();
      RTHeapRep.Unlock ();

      Thread.Pause (update);
    END;
  END Producer;

PROCEDURE SendTypes () =
  VAR t: RT0.TypeDefn;  str: ADDRESS;  n: INTEGER;  p: UNTRACED REF RT0.String;
  BEGIN
    RTHeapRep.Lock ();
      (* describe the types *)
      Send (RT0u.nTypes);
      FOR i := 0 TO RT0u.nTypes-1 DO
        t := RTType.Get (i);
        IF (t.nDimensions > 0)
          THEN Send (-1);          (* variable sized object *)
          ELSE Send (t.dataSize);  (* fixed size *)
        END;
        p := RT0u.type_names + i * ADRSIZE (RT0.String);
        str := p^;
        n := Cstring.strlen (str);
        Send (n);
        Flush ();
        EVAL RTPerfTool.Send (drain, str, n);
      END;
    RTHeapRep.Unlock ();
  END SendTypes;

PROCEDURE Send (i: INTEGER) =
  BEGIN
    IF (cnt >= NUMBER (buf)) THEN Flush () END;
    buf[cnt] := i;
    INC (cnt);
  END Send;

PROCEDURE Flush () =
  BEGIN
    IF (cnt > 0) THEN
      EVAL RTPerfTool.Send (drain, ADR (buf), cnt * BYTESIZE (buf[0]));
      cnt := 0;
    END;
  END Flush;

PROCEDURE SetUpdate (txt: TEXT) =
  VAR n: INTEGER := 0;  ch: INTEGER;
  BEGIN
    IF (txt = NIL) OR Text.Length (txt) = 0 THEN RETURN END;
    FOR i := 0 TO Text.Length(txt)-1 DO
      ch := ORD (Text.GetChar (txt, i)) - ORD ('0');
      IF (ch < 0) OR (9 < ch) THEN RETURN END;
      n := 10 * n + ch;
    END;
    update := MAX (1.0d0, FLOAT (n, LONGREAL));
  END SetUpdate;

PROCEDURE Init () =
  BEGIN
    IF RTPerfTool.Start ("shownew", drain) THEN
      SendTypes ();
      SetUpdate (RTParams.Value ("update"));
      self := Thread.Fork (NEW (Closure));
    END;
  END Init;

BEGIN
END RTHeapInfo.

