(* Copyright (C) 1990, Digital Equipment Corporation.         *)
(* All rights reserved.                                       *)
(* See the file COPYRIGHT for a full description.             *)

(* Last modified on Tue Nov 24 11:22:50 PST 1992 by kalsow    *)
(*      modified on Sat Oct  6 02:01:36 1990 by muller        *)

(* RTIO is a simple low-level interface for I/O on the
   "standard" files.  None of its routines do any locking
   or require any memory allocation.
*)

INTERFACE RTIO;

CONST
  stdin  = 0;
  stdout = 1;
  stderr = 2;

TYPE
  File = INTEGER;

PROCEDURE OpenRead (t: TEXT): File;
(* Open and return a handle suitable for reading the file named "t".
   OpenRead returns -1 if it fails. *)

PROCEDURE OpenWrite (t: TEXT): File;
(* Open and return a handle suitable for writing the file named "t".
   OpenWrite returns -1 if it fails. *)

PROCEDURE Close (f: File);
(* Flush and close the file "f" and release its resources. *)

PROCEDURE EOF (f: File): BOOLEAN;
(* Return "TRUE" if file "f" is at end of file. *)

PROCEDURE GetChar (f: File): CHAR;
(* Return the next character of file "f". *)

PROCEDURE UnGetChar (f: File; c: CHAR);
(* Return "c" to the head of the stream of characters being read
   from file "f".  It's an unchecked runtime error to return more
   than one character between calls to "GetChar". *)

PROCEDURE PutChar (f: File;  c: CHAR);
(* Write "c" on file "f". *)

PROCEDURE PutChars (f: File;  c: UNTRACED REF CHAR;  n: INTEGER);
(* Write characters "c^[0..n-1]" on file "f". *)

PROCEDURE PutText (f: File;  t: TEXT);
(* Write text "t" on file "f". *)

PROCEDURE PutInt (f: File;  i: INTEGER;  width := 0);
(* Write convert integer "i" to characters and write it
   right-justified in a field of "width" characters on file "f". *)

PROCEDURE PutHexa (f: File;  i: INTEGER;  width := 0);
(* Write convert integer "i" to hexidecimal characters and write it
   right-justified in a field of "width" characters on file "f". *)

PROCEDURE Flush (f: File);
(* Flush any buffered characters from "f" to the operating system. *)

END RTIO.
