(* Copyright (C) 1990, Digital Equipment Corporation           *)
(* All rights reserved.                                        *)
(* See the file COPYRIGHT for a full description.              *)

(* Last modified on Fri Feb 18 07:54:57 PST 1994 by kalsow     *)
(*      modified on Wed Jun  2 15:22:58 PDT 1993 by muller     *)

(* The linker generates an inital direct call to this module's
   main body.  All Modula-3 code reached from here. *)

UNSAFE MODULE RTLinker;

IMPORT RT0, RT0u, RTLoader, RTParams, RTProcess, RTHeapRep;
IMPORT RTTypeSRC, RTSignal, RTHooks, RTThreadInit, RTHeapInfo;

VAR init_done := FALSE;

PROCEDURE RunInitPhase (phase: INTEGER) =
  VAR
    n: RT0.ModulePtr;
    m: UNTRACED REF RT0.ModulePtr := info.modules;
  BEGIN
    FOR i := 0 TO info.n_modules - 1 DO
      n := m^;
      IF (n # NIL) AND (n.init # NIL) THEN n.init (phase) END;
      INC (m, ADRSIZE (m^));
    END;
  END RunInitPhase;

BEGIN
  IF NOT init_done THEN
    init_done := TRUE;

    (* spread the linker variables around where they're needed *)
    RT0u.nModules    := info.n_modules;
    RT0u.modules     := info.modules;
    RT0u.nTypes      := info.n_types;
    RT0u.types       := info.types;
    RT0u.type_names  := info.type_names;
    RT0u.alloc_cnts  := info.alloc_cnts;
    RT0u.alloc_bytes := info.alloc_bytes;

    RTHooks.bottom_of_stack := info.bottom_of_stack;
    RTHooks.top_of_stack    := info.top_of_stack;

    (* run each of the init procs to initialize the interface records *)
    RunInitPhase (1);

    (* fix the TypeLink contents *)
    RTTypeSRC.InitLinks ();

    (* run each of the init procs to initialize their types' method lists *)
    RunInitPhase (2);

    (* finish the method lists *)
    RTTypeSRC.InitMethods ();
   
    (* initialize the runtime *)
    RTSignal.InstallHandlers ();
    RTParams.Init ();
    RTLoader.Init ();
    RTHeapRep.Init ();
    RTThreadInit.Init ();
    RTHeapInfo.Init ();

    (* run the module main bodies *)
    RunInitPhase (3);

    (* and be done *)
    RTProcess.Exit (0);
  END;
END RTLinker.


