(* Copyright (C) 1990, Digital Equipment Corporation           *)
(* All rights reserved.                                        *)
(* See the file COPYRIGHT for a full description.              *)

(* Last modified on Tue Dec  7 09:49:20 PST 1993 by kalsow     *)
(*      modified on Fri May 28 14:54:41 PDT 1993 by muller     *)

UNSAFE MODULE RTType EXPORTS RTType, RTTypeSRC;

IMPORT RT0, RT0u, RTMisc, RTModule, RTIO, RTHeapRep, RTLinker;

(*------------------------------------------------ user callable routines ---*)

PROCEDURE MaxTypecode (): Typecode =
  BEGIN
    RETURN RT0u.nTypes - 1;
  END MaxTypecode;

PROCEDURE IsSubtype (a, b: Typecode): BOOLEAN =
  VAR t := Get (b);
  BEGIN
    IF (a >= RT0u.nTypes) THEN BadType (a) END;
    IF (a = 0)            THEN RETURN TRUE END;
    RETURN (t.typecode <= a AND a <= t.lastSubTypeTC);
  END IsSubtype;

PROCEDURE Supertype (tc: Typecode): Typecode =
  VAR t := Get (tc);
  BEGIN
    IF (t.parent = NIL)
      THEN RETURN NoSuchType;
      ELSE RETURN t.parent.typecode;
    END;
  END Supertype;

PROCEDURE IsTraced (tc: Typecode): BOOLEAN =
  VAR t := Get (tc);
  BEGIN
    RETURN t.traced # 0;
  END IsTraced;

PROCEDURE Get (tc: Typecode): RT0.TypeDefn =
  VAR p: UNTRACED REF RT0.TypeDefn := RT0u.types + tc * ADRSIZE (RT0.TypeDefn);
  BEGIN
    IF (tc >= RT0u.nTypes) THEN BadType (tc) END;
    RETURN p^;
  END Get;

PROCEDURE GetNDimensions (tc: Typecode): CARDINAL =
  VAR t := Get (tc);
  BEGIN
    RETURN t.nDimensions;
  END GetNDimensions;

(*----------------------------------------------- TypeLink initialization ---*)

VAR DEBUG := FALSE;

PROCEDURE InitLinks () =
  VAR
    mi   : RT0.ModulePtr;
    t, u : UNTRACED REF RT0.TypeLink;
    defn : RT0.TypeDefn;
  BEGIN
    (* install the exported full revelations *)
    FOR i := 0 TO RT0u.nModules - 1 DO
      mi    := RTModule.Get (i);
      IF (mi = NIL) THEN BadModule (i) END;
      t := mi.type_cells;
      WHILE (t # NIL) DO
        u := t.next;
        defn := FindType (t.type);
        IF (defn = NIL) THEN BadTypeId (mi, t.type) END;
        t.next := defn;
        t.type := defn.typecode;
        t := u;
      END;
    END;
  END InitLinks;

PROCEDURE FindType (id: INTEGER): RT0.TypeDefn =
  VAR
    base : ADDRESS  := RTLinker.info.type_ids;
    lo   : CARDINAL := 0;
    hi   : CARDINAL := RTLinker.info.n_type_ids;
    mid  : CARDINAL; 
    p    : UNTRACED REF INTEGER;
  BEGIN
    WHILE (lo < hi) DO
      mid := (lo + hi) DIV 2;
      p := base + mid * ADRSIZE (INTEGER);
      IF (id < p^)
        THEN hi := mid;
        ELSE lo := mid + 1;
      END;
    END;
    IF (lo > 0) THEN DEC (lo) END;
    p := base + lo * ADRSIZE (INTEGER);
    IF (p^ # id) THEN RETURN NIL END;
    p := RTLinker.info.type_codes + lo * ADRSIZE (INTEGER);
    RETURN Get (p^);
  END FindType;

(*-------------------------------------------- method list initialization ---*)

PROCEDURE InitMethods () =
  BEGIN
    FOR i := 0 TO RT0u.nTypes-1 DO
      FixMethods (Get (i));
    END;

    IF DEBUG THEN
      PutText ("---------------- after type initialization ---------------\n");
      ShowTypes (full := TRUE);
      PutText ("----------------------------------------------------------\n");
    END;

    RTHeapRep.CheckTypes ();
  END InitMethods;

PROCEDURE FixMethods (t: RT0.TypeDefn) =
  VAR m: RT0.MethodSuite;  a, b: UNTRACED REF ADDRESS;
  BEGIN
    m := t.defaultMethods;
    IF (m = NIL) THEN RETURN END;
    <*ASSERT m.typecode = t.typecode*>

    (* inherit my parent's methods *)
    IF (t.parent # NIL) THEN
      a := t.defaultMethods + ADRSIZE (ADDRESS);
      b := t.parent.defaultMethods + ADRSIZE (ADDRESS);
      FOR j := 1 TO t.parent.methodSize DIV BYTESIZE (ADDRESS) - 1 DO
        IF (a^ = NIL) THEN a^ := b^ END;
        INC (a, ADRSIZE (ADDRESS));
        INC (b, ADRSIZE (ADDRESS));
      END;
    END;

    (* initialize any remaining methods to the undefined procedure *)
    a := t.defaultMethods + ADRSIZE (ADDRESS);
    FOR j := 1 TO t.methodSize DIV BYTESIZE (ADDRESS) - 1 DO
      IF (a^ = NIL) THEN a^ := LOOPHOLE (UndefinedMethod, ADDRESS) END;
      INC (a, ADRSIZE (ADDRESS));
    END;
  END FixMethods;

PROCEDURE UndefinedMethod () =
  BEGIN
    RTMisc.FatalError (NIL, 0, "attempted invocation of undefined method");
  END UndefinedMethod;

(*---------------------------------------------------------------- errors ---*)

PROCEDURE BadType (tc: Typecode) =
  BEGIN
    RTMisc.FatalErrorI ("improper typecode: ", tc);
  END BadType;

PROCEDURE BadTypeId (<*UNUSED*> mi: RT0.ModulePtr;  id: INTEGER) =
  BEGIN
    RTMisc.FatalErrorI ("undefined type ", id);
  END BadTypeId;

PROCEDURE BadModule (i: INTEGER) =
  BEGIN
    RTMisc.FatalErrorI ("empty slot in module table ", i);
  END BadModule;

(*---------------------------------------------------- internal debugging ---*)

PROCEDURE ShowTypes (full := TRUE) =
  VAR t: RT0.TypeDefn;
  BEGIN
    PutText ("Here are the types: nTypes = ");
    PutInt  (RT0u.nTypes);
    PutText ("\n");
    FOR i := 0 TO RT0u.nTypes-1 DO
      t := Get (i);
      WHILE (t # NIL) DO
        PutType (t); PutText ("\n");
        IF full THEN
          PutText ("  data   ");
          PutText ("  S= "); PutInt (t.dataSize);
          PutText ("  A= "); PutInt (t.dataAlignment);
          PutText ("  O= "); PutInt (t.dataOffset);
          PutText ("\n");
          IF (t.methodSize # 0) OR (t.methodOffset # 0) THEN
            PutText ("  method ");
            PutText ("  S= ");  PutInt (t.methodSize);
            PutText ("  O= ");  PutInt (t.methodOffset);
            PutText ("\n");
          END;
          IF (t.nDimensions # 0) OR (t.elementSize # 0) THEN
            PutText (" array   ");
            PutText ("  D= ");  PutInt (t.nDimensions);
            PutText ("  S= ");  PutInt (t.elementSize);
            PutText ("\n");
          END;
        END;
      END;
    END;
    RTIO.Flush (RTIO.stderr);
    EVAL ShowTypes; (* to prevent an "unused symbol" warning *)
  END ShowTypes;

PROCEDURE PutType (t: RT0.TypeDefn) =
  BEGIN
    PutText ("[");
    PutAddr (t);

    IF (t # NIL) THEN
      PutText ("  _t");
      PutHex  (t.selfID);

      PutText ("  typecode= ");
      PutIntX (t.typecode, 3);
      IF (t.lastSubTypeTC # 0) THEN
        PutText (" .. ");
        PutIntX (t.lastSubTypeTC, 3);
      END;
    END;

    PutText ("]");
  END PutType;

PROCEDURE PutText (t: TEXT) =
  BEGIN
    RTIO.PutText (RTIO.stderr, t);
  END PutText;

PROCEDURE PutIntX (i, w: INTEGER) =
  BEGIN
    RTIO.PutInt (RTIO.stderr, i, w);
  END PutIntX;

PROCEDURE PutInt (i: INTEGER) =
  BEGIN
    RTIO.PutInt (RTIO.stderr, i);
  END PutInt;

PROCEDURE PutAddr (a: ADDRESS) =
  BEGIN
    PutHex (LOOPHOLE (a, INTEGER));
  END PutAddr;

PROCEDURE PutHex (i: INTEGER) =
  BEGIN
    RTIO.PutHexa (RTIO.stderr, i);
  END PutHex;

BEGIN
END RTType.

