(* Copyright (C) 1994, Digital Equipment Corporation                         *)
(* All rights reserved.                                                      *)
(* See the file COPYRIGHT for a full description.                            *)
(*                                                                           *)
(* Last modified on Wed May 25 15:18:56 PDT 1994 by kalsow                   *)
(*      modified on Fri May  6 13:32:10 PDT 1994 by detlefs                  *)
(*      modified on Tue Jun 16 10:41:32 PDT 1992 by muller                   *)
(*      modified on Sun Mar  1 16:06:32 PST 1992 by meehan                   *)

UNSAFE MODULE RTutils;

IMPORT RTHeapRep, RTType, RT0, RT0u, RTIO, M3toC, RTTypeRec, RTTypeRecSort;

TYPE
  R = REF ARRAY OF RTTypeRec.T;
            
  Visitor = RTHeapRep.RefVisitor OBJECT
              r         : R;
              countSum  := 0;
              sizeSum   := 0
            OVERRIDES
              visit := Walk
            END;

VAR v: Visitor;

PROCEDURE Heap (suppressZeros := FALSE;
                presentation := HeapPresentation.ByTypecode;
                window := LAST(INTEGER)) =
  BEGIN
    Compute ();
    Report (v, suppressZeros, presentation, window)
  END Heap;
  
PROCEDURE NewHeap (suppressZeros := FALSE;
                   presentation := HeapPresentation.ByTypecode;
                   window := LAST(INTEGER)) =
  VAR oldv := v;
  BEGIN
    Compute ();
    Report (Delta (v, oldv), suppressZeros, presentation, window)
  END NewHeap;

PROCEDURE Compute () =
  BEGIN
    v := NEW (Visitor, r := NEW (R, RT0u.nTypes));
    FOR i := 0 TO RT0u.nTypes-1 DO v.r[i].code := i END (* FOR *);
    RTHeapRep.VisitAllRefs (v)
  END Compute;

PROCEDURE Delta (v1, v2: Visitor): Visitor =
  VAR
    v := NEW (Visitor, r := NEW (R, RT0u.nTypes));
  BEGIN
    v.countSum := v1.countSum - v2.countSum;
    v.sizeSum := v1.sizeSum - v2.sizeSum;
    FOR i := 0 TO RT0u.nTypes - 1 DO
      v.r [i].count := v1.r [i].count - v2.r [i].count;
      v.r [i].size := v1.r [i].size - v2.r [i].size
    END;
    RETURN v
  END Delta;
    
PROCEDURE Report (v: Visitor;
                  suppressZeros: BOOLEAN;
                  presentation: HeapPresentation;
                  window: INTEGER) =
  CONST wr = RTIO.stderr;
  VAR nPrinted := 0;
  BEGIN
    CASE presentation OF
    | HeapPresentation.ByTypecode =>
        (*SKIP*)
    | HeapPresentation.ByNumber =>
        RTTypeRecSort.Sort (v.r^, RTTypeRec.CompareCount)
    | HeapPresentation.ByByteCount =>
        RTTypeRecSort.Sort (v.r^, RTTypeRec.CompareSize)
    END (* CASE *);
    RTIO.PutText (wr,
      (* 012345678901234567890123456789012345678901234567890 *)
        "Code   Count   TotalSize  AvgSize  Name\n"
      & "---- --------- --------- --------- --------------------------\n");
    FOR i := 0 TO RT0u.nTypes - 1 DO
      WITH zz =v.r[i], code = zz.code, count = zz.count, size = zz.size DO
        IF (count = 0 AND suppressZeros) OR
           nPrinted >= window THEN (* skip *)
        ELSE
          RTIO.PutInt (wr, code, 4);
          RTIO.PutInt (wr, count, 10);
          RTIO.PutInt (wr, size, 10);
          IF (count = 0)
            THEN RTIO.PutText (wr, "         0");
            ELSE RTIO.PutInt  (wr, size DIV count, 10);
          END;
          RTIO.PutChar (wr, ' ');
          RTIO.PutText (wr, TypecodeName (code));
          RTIO.PutChar (wr, '\n');
          INC(nPrinted);
        END
      END;
    END;
    RTIO.PutText (wr, "     --------- ---------\n    ");
    RTIO.PutInt  (wr, v.countSum, 10);
    RTIO.PutInt  (wr, v.sizeSum, 10);
    RTIO.PutChar (wr, '\n');
  END Report;

PROCEDURE Walk (v : Visitor;
                tc: RTType.Typecode;
                <* UNUSED *> r   : REFANY;
                size: CARDINAL): BOOLEAN =
  BEGIN
    INC (v.r [tc].count);
    INC (v.r [tc].size, size);
    INC (v.countSum);
    INC (v.sizeSum, size);
    RETURN TRUE
  END Walk;

PROCEDURE TypeName (ref: REFANY): TEXT =
  BEGIN
    RETURN TypecodeName (TYPECODE (ref))
  END TypeName;

PROCEDURE TypecodeName (tc: CARDINAL): TEXT =
  BEGIN
    RETURN TypeDefnToName (RTType.Get (tc));
  END TypecodeName;

PROCEDURE TypeDefnToName (t: RT0.TypeDefn): TEXT =
  VAR p: UNTRACED REF RT0.String;  str: ADDRESS;
  BEGIN
    p := RT0u.type_names + t.typecode * ADRSIZE (RT0.String);
    str := p^;
    RETURN M3toC.CopyStoT (str);
  END TypeDefnToName;

(***************************************************
PROCEDURE TypeDefnToName (t: RT0.TypeDefn): TEXT =
  <*FATAL Convert.Failed*>
  CONST N = BITSIZE (INTEGER) + 2;
  VAR len: INTEGER;  buf: ARRAY [0..N-1] OF CHAR;
  BEGIN
    buf [0] := '_';
    buf [1] := 't';
    len := Convert.FromUnsigned (SUBARRAY (buf, 2, N-2), t.selfID, 16);
    RETURN Text.FromChars (SUBARRAY (buf, 0, len+2));
  END TypeDefnToName;
***********************************************************)

BEGIN 
  v := NEW (Visitor, r := NEW (R, RT0u.nTypes))
END RTutils.
