(* Copyright (C) 1989, Digital Equipment Corporation *)
(* All rights reserved. *)
(* See the file COPYRIGHT for a full description. *)
(* Last modified on Thu Jan 28 10:45:24 PST 1993 by mjordan *)
(*      modified on Fri Dec  4 17:35:53 PST 1992 by mcjones *)

UNSAFE MODULE DatePosix EXPORTS Date;

IMPORT Thread, Time, M3toC, Utime, TimePosix;

VAR mu := NEW(Thread.Mutex);
(* Protect the static storage used by "Utime.localtime" and "Utime.gmtime"
   (see the localtime(3) manpage). *)

REVEAL TimeZone = BRANDED OBJECT METHODS fromTime(t: Time.T): T END;

PROCEDURE FromTimeLocal(<*UNUSED*> z: TimeZone; t: Time.T): T =
  VAR tv: Utime.struct_timeval;
  BEGIN
    tv := TimePosix.ToUtime(t);
    LOCK mu DO
      RETURN DateFromStruct_tm(Utime.localtime(ADR(tv.tv_sec)))
    END
  END FromTimeLocal;

PROCEDURE FromTimeUTC(<*UNUSED*> z: TimeZone; t: Time.T): T =
  VAR tv: Utime.struct_timeval;
  BEGIN
    tv := TimePosix.ToUtime(t);
    LOCK mu DO
      RETURN DateFromStruct_tm(Utime.gmtime(ADR(tv.tv_sec)))
    END
  END FromTimeUTC;

PROCEDURE DateFromStruct_tm(tm: Utime.struct_tm_star): T =
  BEGIN
    RETURN T{
      year := tm.tm_year + 1900,
      month := VAL(tm.tm_mon, Month),
      day := tm.tm_mday,
      hour := tm.tm_hour,
      minute := tm.tm_min,
      second := tm.tm_sec,
      offset := tm.tm_gmtoff,
      zone := M3toC.CopyStoT(tm.tm_zone),
      weekDay := VAL(tm.tm_wday, WeekDay)
    }
  END DateFromStruct_tm;

PROCEDURE FromTime(t: Time.T; z: TimeZone := NIL): T =
  BEGIN
    IF z = NIL THEN z := Local END;
    RETURN z.fromTime(t)
  END FromTime;

PROCEDURE ToTime(READONLY d: T): Time.T RAISES {Error} =
  VAR
    tm: Utime.struct_tm;
    time: Utime.time_t;
    t: Time.T;
  BEGIN
    tm.tm_sec := d.second;
    tm.tm_min := d.minute;
    tm.tm_hour := d.hour;
    tm.tm_mday := d.day;
    tm.tm_mon := ORD(d.month);
    tm.tm_year := d.year - 1900;
    (* tm.tm_wday ignored *)
    tm.tm_isdst := 0; (* tell mktime that DST is not in effect *)
    tm.tm_gmtoff := d.offset;
    (* tm_zone ignored *)
    time := Utime.mktime(ADR(tm));
    IF time = -1 THEN RAISE Error END;
    t := FLOAT(time, LONGREAL);
    RETURN t
  END ToTime;

BEGIN
  Local := NEW(TimeZone, fromTime := FromTimeLocal);
  UTC := NEW(TimeZone, fromTime := FromTimeUTC)
END DatePosix.
